import * as tsplus_module_1 from "@effect/core/io/Effect/operations/map";
import * as tsplus_module_2 from "@effect/core/io/Effect/operations/zip";
import * as tsplus_module_3 from "@tsplus/stdlib/collections/Chunk/definition";
import * as tsplus_module_4 from "@effect/core/io/Effect/operations/sync";
import * as tsplus_module_5 from "@effect/core/io/Effect/operations/unit";
import * as tsplus_module_6 from "@tsplus/stdlib/collections/SortedSet/add";
import * as tsplus_module_7 from "@tsplus/stdlib/collections/SortedSet/remove";
export const unsafeTrack = unsafeTrack_1;
export const fromEffect = fromEffect_1;
export const SupervisorURI = /*#__PURE__*/Symbol.for("@effect/core/io/Supervisor");
/**
 * A `Supervisor<A>` is allowed to supervise the launching and termination of
 * fibers, producing some visible value of type `A` from the supervision.
 *
 * @tsplus type effect/core/io/Supervisor
 * @tsplus companion effect/core/io/Supervisor.Ops
 */

export class Supervisor {
  /**
   * Maps this supervisor to another one, which has the same effect, but whose
   * value has been transformed by the specified function.
   */
  map(f) {
    return new ProxySupervisor(this, () => tsplus_module_1.map(f)(this.value));
  }
  /**
   * Returns a new supervisor that performs the function of this supervisor, and
   * the function of the specified supervisor, producing a tuple of the outputs
   * produced by both supervisors.
   */


  zip(right) {
    return new Zip(this, right);
  }

}
export class ProxySupervisor extends Supervisor {
  constructor(underlying, value0) {
    super();
    this.underlying = underlying;
    this.value0 = value0;
  }

  get value() {
    return this.value0();
  }

  onStart(environment, effect, parent, fiber) {
    this.underlying.onStart(environment, effect, parent, fiber);
  }

  onEnd(value, fiber) {
    this.underlying.onEnd(value, fiber);
  }

  onEffect(fiber, effect) {
    this.underlying.onEffect(fiber, effect);
  }

  onSuspend(fiber) {
    this.underlying.onSuspend(fiber);
  }

  onResume(fiber) {
    this.underlying.onResume(fiber);
  }

}
export class Zip extends Supervisor {
  constructor(left, right) {
    super();
    this.left = left;
    this.right = right;
  }

  get value() {
    return tsplus_module_2.zip(this.right.value)(this.left.value);
  }

  onStart(environment, effect, parent, fiber) {
    this.left.onStart(environment, effect, parent, fiber);
    this.right.onStart(environment, effect, parent, fiber);
  }

  onEnd(value, fiber) {
    this.left.onEnd(value, fiber);
    this.right.onEnd(value, fiber);
  }

  onEffect(fiber, effect) {
    this.left.onEffect(fiber, effect);
    this.right.onEffect(fiber, effect);
  }

  onSuspend(fiber) {
    this.left.onSuspend(fiber);
    this.right.onSuspend(fiber);
  }

  onResume(fiber) {
    this.left.onResume(fiber);
    this.right.onResume(fiber);
  }

}
export class Track extends Supervisor {
  constructor() {
    super(...arguments);
    this.fibers = new Set();
  }

  get value() {
    return tsplus_module_4.sync(() => tsplus_module_3.from(this.fibers));
  }

  onStart(_environment, _effect, _parent, fiber) {
    this.fibers.add(fiber);
  }

  onEnd(_value, fiber) {
    this.fibers.delete(fiber);
  }

  onEffect(_fiber, _effect) {//
  }

  onSuspend(_fiber) {//
  }

  onResume(_fiber) {//
  }

}
export class Const extends Supervisor {
  constructor(effect) {
    super();
    this.effect = effect;
  }

  get value() {
    return this.effect;
  }

  onStart(_environment, _effect, _parent, _fiber) {//
  }

  onEnd(_value, _fiber) {//
  }

  onEffect(_fiber, _effect) {//
  }

  onSuspend(_fiber) {//
  }

  onResume(_fiber) {//
  }

}
/**
 * Creates a new supervisor that tracks children in a set.
 *
 * @tsplus static effect/core/io/Supervisor.Ops unsafeTrack
 */

function unsafeTrack_1() {
  return new Track();
}
/**
 * Creates a new supervisor that tracks children in a set.
 *
 * @tsplus static effect/core/io/Supervisor.Ops track
 */


export const track = /*#__PURE__*/tsplus_module_4.sync(unsafeTrack_1);
/**
 * Creates a new supervisor that constantly yields effect when polled
 *
 * @tsplus static effect/core/io/Supervisor.Ops fromEffect
 */

function fromEffect_1(effect) {
  return new Const(effect);
}
/**
 * A supervisor that doesn't do anything in response to supervision events.
 *
 * @tsplus static effect/core/io/Supervisor.Ops none
 */


export const none = /*#__PURE__*/fromEffect_1(tsplus_module_5.unit);

class FibersIn extends Supervisor {
  constructor(ref) {
    super();
    this.ref = ref;
  }

  get value() {
    return tsplus_module_4.sync(() => this.ref.get);
  }

  onStart(_environment, _effect, _parent, fiber) {
    this.ref.set(tsplus_module_6.add(fiber)(this.ref.get));
  }

  onEnd(_value, fiber) {
    this.ref.set(tsplus_module_7.remove(fiber)(this.ref.get));
  }

  onEffect(_fiber, _effect) {//
  }

  onSuspend(_fiber) {//
  }

  onResume(_fiber) {//
  }

}
/**
 * Creates a new supervisor that tracks children in a set.
 *
 * @tsplus static effect/core/io/Supervisor.Ops fibersIn
 */


export function fibersIn(ref) {
  return tsplus_module_4.sync(() => new FibersIn(ref));
}
//# sourceMappingURL=definition.mjs.map